{
  CMF Player - (c) vIRtECH 1996-1997 Ansgar Scherp

  Version 3.1a - some major and minor bugfixes and some asm speed - up.

  Source-Release Version

  PLAYING CMF-FILES WITHOUT USING THE SBFMDRV FROM CREATIVE LABS (R).
      FOR PROTECTED MODE and REAL MODE including FULL SOURCE CODE!

  (c) vIRTUAL tECHNOLOGIES 1996

  Based on the source by :
    Carl Eric Codere ( codc01@gel.usherb.ca )
    http://www-edu.gel.usherb.ca/codc01

  Special thanks to you.

  Distribition via any media is explicit allowed.

  DISCLAIMER:

  We give no warranty for this code, because it was firstly coded for us,
  and not for the public.
  But after lot of people asking us for releasing the source, we decided,
  to make it now available.
  But however, I really don't think that there is a bug in this unit.

  You're not allowed to re-release the source in a modified version of it.
  However, you can use it as it is for any (even commercial) projects, but
  you have to give creedits to the authors.

  for more information contact us:

    Ansgar Scherp
    scherp.ansgar@informatik.uni-oldenburg.de

    http://www.informatik.uni-oldenburg.de/~virtech/

    Mailbox : + 49 (0) 4441 / 851887
    Hotline : + 49 (0) 4441 / 851717 (english, german)

  Please report, if you find any bugs in this source code.

}

{$S-,F+}
unit _CMFPLAY;


interface

uses _AdLib, DOS;

type
  TCMFError =
    ( CMFInvalidFile , CMFInvalidVersion );
  { *** Some Channel and Header Information }
  TChannels =
    Array[0..15] of Byte;
  { *** CMF - Headerinformation *** }
  TCMFHeader =
    Record
      CMFFileID        : Array[0..3] of Char; { should be 'CTMF'            }
      Version          : Word; { LSB = Minor Version, MSB = Major Version   }
      InstrumentOffset : Word; { Offset of file into Instrument block       }
      MusicOffset      : Word; { Offset of file into music block            }
      TicksPerQuartet  : Word; { Clock ticks per quarter note
                                 Calculation := ( TicksPerQuarter * Tempo )
                                 div 60 sec.                                }
      TicksPerSecond   : Word; { Clock Ticks per second                     }
      TitleOffset      : Word; { Offset of file to title                    }
      AuthorOffset     : Word; { Offset of file to author name              }
      NotaOffset       : Word; { Offset of file to notations                }
      { For each voice, tells if voice is used or not:
        0 = Unused voice; 1 = Used voice                                    }
      Channels         : TChannels; { For all 16 midi CHANNELS Offset 20-35 }
      Instruments      : Word;      { Number of instruments used-1    36-37 }
      Tempo            : Word;      { Offset 38-39 }
    end;


 { *** CMF-Datei Instrumenten Record *** }
 TCMFInstrument =
   Record
     ModulatorChar      : Byte; { Modulator Characteristic }
     CarrierChar        : Byte; { Carrier Characteristic   }
     ModScalingOutput   : Byte;
     CarScalingOutput   : Byte;
     ModAttackDelay     : Byte;
     CarAttackDelay     : Byte;
     ModSustainRelease  : Byte;
     CarSustainRelease  : Byte;
     ModWaveSelect      : Byte;
     CarWaveSelect      : Byte;
     FeedBackConnection : Byte;
     Reserved           : Array[1..5] of Byte;
   end;

var
  CurrentCMFPos : Word;       { Current Position in song  }
  CMFPlaying    : boolean;    { is the song Playing ??? }
  CMFRepeat     : boolean;    { repeat song at end of file ??? }
  Header        : TCMFHeader; { Header of the CMF-File }
  Channels      : TChannels;  { Channels in use }
  CMFSpeed      : real;       { current playing speed = 100% * CMFSpeed }
  CMFErrorValue : TCMFError;

{ Loads specified CMF-File to memory and starts playing }
procedure CMFInit(AFileName: string );
{ Stops playing and closes the CMF-Song }
procedure CMFClose;

{ Toggles between Pausing and Playing then CMF-Song }
procedure CMFPausePlay;
{ stopps playing CMFFile and sets song position to beginning }
procedure CMFStop;
{ checks if an Adlib-SoundCard is installed }
function AdlibBoardInstalled : boolean;
{ verifies the CMF-Song }
function VerifyCMFFile : boolean;

implementation

procedure CMFInterruptHandler; interrupt; forward;

const
  { you may modify this value for memory - optimizing /
    one time it'll be all dynamic ... }
  MaximumCMFFileSize = 65535;

  TimerIntr        = 8;       { use TimerInterrupt for polling }
  ProgIntTimerFreq = $1234DD;


type
  TCMFFile = Array[1..MaximumCMFFileSize] of Byte;
  PCMFFile = ^TCMFFile;

var
  BIOSTimerHandler    : procedure;
  ClockTicks,
  ClockCounter        : longint;
  CopiedValue         : Byte;
  DelayTime           : Word;       { current delay time }
  CMFFile             : PCMFFile;   { pointer to current song data }
  InstrTable          : Array[0..255] of TCMFInstrument;

{ passes function BoardInstalled from _AdLib }
function AdlibBoardInstalled:Boolean;
begin
  AdlibBoardInstalled := BoardInstalled;
end;

{ turns all notes off }
procedure TurnAllNotesOff;
var
  Voice : byte;
begin
  for Voice := 0 to 15 do
    begin
      NoteOff( Voice, CurrentPlayingNote[ Voice ] );
      if Voice < 13 then SetVolume( Voice, 0 );
    end;
end;

{ toggles between playing and pausing the song }
procedure CMFPausePlay;
var
  Voice : byte;
begin
  if CMFPlaying then TurnAllNotesOff;
  CMFPlaying := not CMFPlaying;
end;

{ stops cmf-song if playing and resets song position at beginning }
procedure CMFStop;
begin
  CMFPlaying := false;
  TurnAllNotesOff;
  CurrentCMFPos := 1;
end;

{ called when there is an invalid music file format }
Procedure CMFError( ErrorValue: TCMFError );
Begin
 Case ErrorValue of
   CMFInvalidFile    :
     begin
       WriteLn('CMF-ERROR: Invalid File Format');
       Halt(1);
     end;
   CMFInvalidVersion :
     begin
       WriteLn('CMF-ERROR: Invalid CMF Version');
       Halt(2);
     end;
   else
     begin
       WriteLn('Unknown CMF-ERROR!');
       Halt(3);
     end;
 end;
end;


{ *** MAIN ROUTINES *** }

procedure SetPITimer(TimerHandler : pointer; frequency : word);
begin
  { Do some initialization }
  ClockTicks := 0;
  ClockCounter := ProgIntTimerFreq div frequency;

  { Store the current BIOS handler and set up our own }
  GetIntVec( TimerIntr, @BIOSTimerHandler);
  SetIntVec( TimerIntr, TimerHandler);

  { Set the PIT channel 0 frequency }
  Port[$43] := $34;
  Port[$40] := Byte(ClockCounter mod 256);
  Port[$40] := Byte(ClockCounter div 256);
end;


procedure CleanUpPITimer;
begin
  { Restore the normal clock frequency }
  Port[$43] := $36;
  Port[$40] := 0;
  Port[$40] := 0;
  { Restore the normal ticker handler }
  SetIntVec( TimerIntr, @BIOSTimerHandler);
end;

procedure CMFInit(AFileName: String );
VAR
  i : Byte;
  F : file;
Begin
  FileMode := 0;

  CMFSpeed := 1; { set speed to 100% }

  i := 0;

  Assign(F, AFileName );
  Reset( F, 1 );
  BlockRead(F, Header, SizeOf(Header));

  if (not VerifyCMFFile ) then CMFError( CMFInvalidFile );

  New( CMFFile );
  AdLibInit;
  Seek(F, Header.InstrumentOffset);

  { reading instruments }
  repeat
    BlockRead(F, InstrTable[i], SizeOf(TCMFInstrument));
    inc(i);
  until i = Header.Instruments;

  Channels := Header.Channels;

  Seek( F,Header.MusicOffset);
  BlockRead(F,  CMFFile ^, FileSize(F) - FilePos(F) );

  Close(F);

  CurrentCMFPos := 1;

  SetPITimer( @CMFInterruptHandler ,
              Header.TicksperSecond + Header.TicksperSecond div 4 );
end;


{ Sets the Instruments passed as a parameter to the specified
  music card voice.}
Procedure SetVoice( Voice:Byte; VAR Instr: TCMFInstrument);
var
  TablePModulatorHelp : byte;
Begin
  if (Voice < 9) and (not Percussive) OR ((Percussive)
   { In Melodic mode call this                         }
   { In percussive mode, calls this if Bass drum modes }
   { or if a melodic voice is used.                    }
     and (Voice < 7)) then
    begin
      SetWaveSelect( Voice, Modulator, Instr.ModWaveSelect );
      SetWaveSelect( Voice, Carrier,   Instr.CarWaveSelect );
      SetWaveChar(   Voice, Modulator, Instr.ModulatorChar );
      SetWaveChar(   Voice, Carrier,   Instr.CarrierChar );
      AllScalingOutput( Voice, Modulator, Instr.ModScalingOutput );
      AllScalingOutput( Voice, Carrier,   Instr.CarScalingOutput );
      { Save the scaling output for the second operator }
      { this is the type which controls the volume of   }
      { the voice.                                      }
      CarrierScalingLevel[ Voice ]   := Instr.CarScalingOutput and $C0;
      ModulatorScalingLevel[ Voice ] := Instr.ModScalingOutput and $C0;
      Algorithm[ Voice ] := Boolean( Instr.FeedBackConnection and $01 );
      MaxVolume[ Voice ] := Instr.CarScalingOutput and $3f;
      MaxVolume[ Voice ] := Byte( MaxVolume[Voice] - $3F );
      MaxVolume[ Voice ] := Byte( 0 - MaxVolume[Voice] );
      AllAttackDecay( Voice, Modulator, Instr.ModAttackDelay );
      AllAttackDecay( Voice, Carrier, Instr.CarAttackDelay   );
      AllFeedBack(   Voice, Instr.FeedBackConnection );
      AllSusRelease( Voice, Modulator, Instr.ModSustainRelease );
      AllSusRelease( Voice, Carrier,   Instr.CarSustainRelease );
    end
  else
  { This is a normal percussive voice }
  { Bug fix: Percussive mode music instruments in CMF file still use }
  { carrier level for normal output level calculation.               }
    begin
      TablePModulatorhelp := TablePModulator[Voice];
      WriteAdlibPort( TablePModulatorHelp + $20, Instr.ModulatorChar     );
      WriteAdlibPort( TablePModulatorHelp + $40, Instr.ModScalingOutput  );
      WriteAdlibPort( TablePModulatorHelp + $60, Instr.ModAttackDelay    );
      WriteAdlibPort( TablePModulatorHelp + $80, Instr.ModSustainRelease );
      WriteAdlibPort( Voice                  + $C0, Instr.FeedbackConnection);
      WriteAdlibPort( TablePModulatorHelp + $E0, Instr.ModWaveSelect     );
      CarrierScalingLevel[ Voice ]   := Instr.CarScalingOutput and $C0;
      ModulatorScalingLevel[ Voice ] := Instr.ModScalingoutput and $C0;
      Algorithm[ SnareDrumVoice ]    :=
                                   Boolean(Instr.FeedBackConnection and $01);
      MaxVolume[ Voice ] := Instr.CarScalingOutput and $3f;
      MaxVolume[ Voice ] := Byte(MaxVolume[Voice] - $3F);
      MaxVolume[ Voice ] := Byte(0 - MaxVolume[Voice]);
    end;
end;


{ Verifies if the CMFFile is a valid format.}
Function VerifyCMFFile : Boolean;
begin
 VerifyCMFFile := TRUE;
 if Header.CMFFileID <> 'CTMF' then
   Begin
     VerifyCMFFile  := FALSE;
     CMFErrorValue  := CMFInvalidFile;
   end;
end;


procedure CMFClose;
begin
  CleanUpPITimer;
  AdLibDone;
  AdLibDone;
  if CMFFile <> Nil then Dispose( CMFFile );
end;


procedure CalculateDelayTime(VAR Index: Byte);
Var
  LocalIndex: Byte;
  i: Byte;

  byteValues: Array[0..3] of Byte;
  Power: Real;
  ActData: Byte;

  Help1, Help2 : word;

Begin
  asm
  {For i:=0 to 3 do ByteValues[i] := 0;
  LocalIndex := 0;
  ActData := 128;}
    mov word ptr ByteValues[0], 0
    mov word ptr ByteValues[2], 0
    mov LocalIndex ,   0
    mov ActData    , 128
  end;

  WHILE (ActData and $80) = $80 DO
    BEGIN
      ActData :=  CMFFile ^[ CurrentCMFPos + Index + 1 ];
      ByteValues[LocalIndex] := ActData;
      inc(Index);
      asm
        inc LocalIndex
      end;
    END;

  asm
    mov DelayTime,0
  end;
  FOR i:= 0 to (LocalIndex-1) do
    Begin
      { Power := Exp(i * Ln(128));
        DelayTime := DelayTime +
                 (Trunc(Power)*(ByteValues[(LocalIndex-1)-i] and $7F));}
      { REPLACED BY: }
{      DelayTime := DelayTime + (ByteValues[i] and $7F)
                   SHL (7*(LocalIndex-i-1));}

      Help1 := ByteValues[i];
      asm
        mov ax, Help1
        and ax, $7F
        add ax, DelayTime
        mov Help1, ax

        mov ax, word( LocalIndex )
        dec ax
        sub ax, word(i)
        mov bx, 7
        mul ax
        mov Help2, ax
      end;
      DelayTime := Help1 SHL Help2;

    end;
  DelayTime := trunc( DelayTime / CMFSpeed );
end;

{ many more has to move into asm }
Procedure ProcessCMFEvent;

Var
 Volume,
 Index,
 Note,
 Voice,
 CurrentSongPosByte : byte;

Begin
  if CMFPlaying = false then exit;
  if DelayTime > 1 then begin Dec( DelayTime ); Exit; end;

  asm
    mov Volume , 0
    mov Note   , 0
    mov Voice  , 0
  end;

  CurrentSongPosByte := CMFFile^[CurrentCMFPos];

  Case CurrentSongPosByte of
    { Note Number Off // Main CASE }
    $80..$8F :
      Begin
        {CopiedValue := CurrentSongPosByte;
        Voice       := CurrentSongPosByte and $0F;}
        asm
          mov ah, CurrentSongPosByte
          mov CopiedValue, ah
          and ah, $0F
          mov Voice, ah
        end;
        { CMF channels 12 to 15 are the equivalent of voices }
        { 7,8,9,10                                           }
        if Voice > 9 then dec(Voice,5);
        Note   := CMFFile^[ CurrentCMFPos + 1 ];
        Volume := CMFFile^[ CurrentCMFPos + 2 ];
        Index:=2;
        if Volume = 0 then SetVolume(Voice, Volume);
        NoteOff( Voice, Note );
        CalculateDelayTime(Index);
        asm
          {Inc(CurrentCMFPos, Index + 1);}
          xor ax, ax
          mov al, Index
          inc ax
          add CurrentCMFPos, ax
        end;
      end;
    { Note Number on // Main CASE }
    $90..$9F:
      Begin
        asm
          mov Index, 2;
        end;
        CopiedValue:=CurrentSongPosByte;
        Voice  := CMFFile^[CurrentCMFPos] and $0F;
        { CMF channels 12 to 15 are the equivalent of voices }
        { 7,8,9,10                                           }
        if Voice > 9 then Voice := Voice - 5;
        Note   := CMFFile^[ CurrentCMFPos + 1 ];
        Volume := CMFFile^[ CurrentCMFPos + 2 ];
        if Volume = 0 then
          NoteOff(Voice, note)
          else
            Begin
              SetVolume(Voice, Volume);
              NoteOn(Voice, Note);
            end;
        CalculateDelayTime(Index);
        asm
          {Inc(CurrentCMFPos, Index + 1);}
          xor ax, ax
          mov al, Index
          inc ax
          add CurrentCMFPos, ax
        end;
      end;
    { Check if end of song // Main CASE }
    $FF :
      Begin
        if (CMFFile^[CurrentCMFPos + 1] = $2f) and (CMFFile^[CurrentCMFPos + 2] = $00) then
          Begin
            Volume:=CMFFile^[ CurrentCMFPos + 2 ];
            {Playing:=False;}
            if CMFRepeat then
                CurrentCMFPos:=1
              else
                begin
                  CMFPlaying := false;
                  CurrentCMFPos:=1
                end;
          end;
        end;
    { Controller Change // Main CASE }
    $B0..$BF:
      Begin
        CopiedValue := CurrentSongPosByte;
        Case CMFFile^[ CurrentCMFPos + 1 ] of
        $66: { Music Marker }
          Begin
            asm
              mov Index,2
            end;
            CalculateDelayTime( Index );
            asm
              {Inc(CurrentCMFPos, Index + 1);}
              xor ax, ax
              mov al, Index
              inc ax
              add CurrentCMFPos, ax
            end;
          end;
        $67: { Specify which mode 1 = PERCUSSIVE }
          Begin
            asm
              mov Index,2
            end;
            SetPercussive(Boolean(CMFFile^[CurrentCMFPos + 2]));
            CalculateDelayTime(Index);
            asm
              {Inc(CurrentCMFPos, Index + 1);}
              xor ax, ax
              mov al, Index
              inc ax
              add CurrentCMFPos, ax
            end;
          end;
        $68: { All note upward }
          Begin
            asm
              mov Index,2
            end;
            Voice := CurrentSongPosByte and $0F;
            { CMF channels 12 to 15 are the equivalent of voices }
            { 7,8,9,10                                           }
            if Voice > 9 then Voice := Voice - 5;
            SemiToneUp(Voice, Shortint(CMFFile^[CurrentCMFPos + 2]));
            CalculateDelayTime(Index);
            asm
              {Inc(CurrentCMFPos, Index + 1);}
              xor ax, ax
              mov al, Index
              inc ax
              add CurrentCMFPos, ax
            end;
          end;
        $69: { All notes downward }
          Begin
            Voice := CurrentSongPosByte and $0F;
            { CMF channels 12 to 15 are the equivalent of voices }
            { 7,8,9,10                                           }
            if Voice > 9 then Voice := Voice - 5;
            asm
              mov Index,2
            end;
            CalculateDelayTime(Index);
            SemiToneDown(Voice, Shortint(CMFFile^[CurrentCMFPos + 2]));
            asm
              {Inc(CurrentCMFPos, Index + 1);}
              xor ax, ax
              mov al, Index
              inc ax
              add CurrentCMFPos, ax
            end;
          end;
        else
          begin
            asm
              mov Index,2
            end;
            CalculateDelayTime( Index );
            asm
              {Inc(CurrentCMFPos, Index + 1);}
              xor ax, ax
              mov al, Index
              inc ax
              add CurrentCMFPos, ax
            end;
          end;
        end; { of SUB CASE}
      end;
    { Change to musical instrument nn // Main CASE }
    $C0..$CF:
      Begin
        Voice := CurrentSongPosByte and $0F;
        { CMF channels 12 to 15 are the equivalent of voices }
        { 7,8,9,10                                           }
        if Voice > 9 then Voice := Voice - 5;
        CopiedValue := CMFFile^[CurrentCMFPos];
        { Do A NoteOff first with any note number }
        NoteOff(Voice, $60);
        SemiToneCurrent[Voice] := 0;
        SetVoice(Voice, InstrTable[CMFFile^[CurrentCMFPos + 1]]);
        asm
          mov Index,1
        end;
        CalculateDelayTime(Index);
        asm
          {Inc(CurrentCMFPos, Index + 1);}
          xor ax, ax
          mov al, Index
          inc ax
          add CurrentCMFPos, ax
        end;
      end;
    $00: asm Inc CurrentCMFPos end;
    else
      { A repeat MIDI EVENT }
      Begin
        Case CopiedValue of
          { Note Number Off }
          $80..$8F:
            Begin
              Voice := CopiedValue and $0F;
              { CMF channels 12 to 15 are the equivalent of voices }
              { 7,8,9,10                                           }
              if Voice > 9 then Voice := Voice - 5;
              Note   := CurrentSongPosByte;
              Volume := CMFFile^[CurrentCMFPos + 1];
              asm
                mov Index,1
              end;
              SetVolume(Voice, Volume);
              NoteOff(Voice, Note);
              CalculateDelayTime(Index);
              asm
                {Inc(CurrentCMFPos, Index + 1);}
                xor ax, ax
                mov al, Index
                inc ax
                add CurrentCMFPos, ax
              end;
            end;
          { Note Number on }
          { Open Voice 1,2,3,4,5,6,7,8 or 9 }
          $90..$9F:
            Begin
              asm
                mov Index,1
                {Voice := CopiedValue and $0F;}
                mov ah, CopiedValue
                and ah, $0F
                mov Voice, ah
              end;
              { CMF channels 12 to 15 are the equivalent of voices }
              { 7,8,9,10                                           }
              if Voice > 9 then Voice := Voice - 5;
              Volume := CMFFile^[CurrentCMFPos + 1];
              Note   := CurrentSongPosByte;
              SetVolume(Voice, Volume);
              if Volume = 0 then
                 NoteOff(Voice, Note)
               else
                 NoteOn(Voice, Note);
              CalculateDelayTime(Index);
              asm
                {Inc(CurrentCMFPos, Index + 1);}
                xor ax, ax
                mov al, Index
                inc ax
                add CurrentCMFPos, ax
              end;
            end;
          { controller change }
          $B0..$BF:
            Begin
              Case CurrentSongPosByte of
                $66: { Music Marker }
                  Begin
                    asm
                      mov Index,1
                    end;
                    CalculateDelayTime(Index);
                    asm
                      {Inc(CurrentCMFPos, Index + 1);}
                      xor ax, ax
                      mov al, Index
                      inc ax
                      add CurrentCMFPos, ax
                    end;
                  end;
                $67: { Specify which mode }
                  Begin
                    asm
                      mov Index,1
                    end;
                    SetPercussive(Boolean(CMFFile^[CurrentCMFPos + 1]));
                    CalculateDelayTime(Index);
                    asm
                      {Inc(CurrentCMFPos, Index + 1);}
                      xor ax, ax
                      mov al, Index
                      inc ax
                      add CurrentCMFPos, ax
                    end;
                  end;
                $68:
                  Begin
                    asm
                      {Voice := CopiedValue and $0F;}
                      mov ah, CopiedValue
                      and ah, $0F
                      mov Voice, ah
                    end;
                    { CMF channels 12 to 15 are the equivalent of voices }
                    { 7,8,9,10                                           }
                    if Voice > 9 then Voice := Voice -5;
                    SemiToneUp(Voice,Shortint(CMFFile^[CurrentCMFPos + 1]));
                    CalculateDelayTime(Index);
                    asm
                      {Inc(CurrentCMFPos, Index + 1);}
                      xor ax, ax
                      mov al, Index
                      inc ax
                      add CurrentCMFPos, ax
                    end;
                  end;
                $69:
                  Begin
                    asm
                      mov Index,1;
                      {Voice := CopiedValue and $0F;}
                      mov ah, CopiedValue
                      and ah, $0F
                      mov Voice, ah
                    end;
                    if Voice > 9 then Voice := Voice -5;
                    SemiToneDown(Voice, Shortint(CMFFile^[CurrentCMFPos + 1]));
                    CalculateDelayTime(Index);
                    asm
                      {Inc(CurrentCMFPos, Index + 1);}
                      xor ax, ax
                      mov al, Index
                      inc ax
                      add CurrentCMFPos, ax
                    end;
                  end;
                else
                  begin
                    asm
                      mov Index, 1
                    end;
                    CalculateDelayTime( Index );
                    asm
                      {Inc(CurrentCMFPos, Index + 1);}
                      xor ax, ax
                      mov al, Index
                      inc ax
                      add CurrentCMFPos, ax
                    end;
                  end;
              end; { SUB2 CASE }
            end;
          { Change to musical instrument nn }
          $C0..$CF:
            Begin
              asm
                {Voice := CopiedValue and $0F;}
                mov ah, Index
                and ah, $0F
                mov Voice, ah
              end;
              { CMF channels 12 to 15 are the equivalent of voices }
              { 7,8,9,10                                           }
              if Voice > 9 then dec( Voice , 5);
              NoteOff(Voice, $60);
              SemiToneCurrent[Voice] := 0;
              SetVoice(Voice, InstrTable[CurrentSongPosByte]);
              asm
                mov Index,1
              end;
              CalculateDelayTime( Index );
              asm
              {Inc(CurrentCMFPos, Index + 1);}
                xor ax, ax
                mov al, Index
                inc ax
                add CurrentCMFPos, ax
              end;
            end;
          else WriteLn('MIDI FORMAT ERROR At Position:',CurrentCMFPos);
        end; { end of ELSE-CASE }
      end; { end of else in MAIN CASE}
  end; {MAIN CASE}
end; {PROC}

{$F+}
procedure CMFInterruptHandler;
var
  Wert : Longint;
begin
  asm
    call ProcessCMFEvent;
  end;
  { adjust the count of clock ticks }
  ClockTicks := ClockTicks + ClockCounter;

{  asm
    mov ax, word ptr ClockTicks + 0
    mov bx, word ptr ClockTicks + 2

    add ax, word ptr ClockCounter + 0
    add bx, word ptr ClockCounter + 2

    mov word ptr ClockTicks + 0, ax
    mov word ptr ClockTicks + 2, bx
  end;}

  { has the BIOS handler to do his work??? }
  if ClockTicks >= $10000 then
    begin
      { Yep! So adjust the count and call the BIOS handler }
      ClockTicks := ClockTicks - $10000;
      asm
        pushf
        call BIOSTimerHandler
      end;
    end
  else
 { no, so make knowledge of the interrupt }
 asm
   mov al, $20
   mov dx, $20
   out dx, al
 end;
end;
{$F-}

begin
  { init start-up values }
  CMFPlaying := false;
  CMFRepeat  := true;
end.


